/*-
 * Copyright (c) 2000 Brian Fundakowski Feldman
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $FreeBSD$
 *
 */

#include <sys/types.h>
#include <sys/cdrio.h>

#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "tracks.h"

int read_cd(int fd, struct tracks **rettracks, int numonly);
int getrange(char *s, unsigned int range[2]);

int
main(int argc, char **argv) {
	struct tracks *tracks;
	const char *device = "/dev/acd0c";
	const char *outputdir = ".";
	int ch, fd, debug = 0, numonly = 0, psx = 0, toconly = 0;
	unsigned int trackrange[2] = { 0, ~0u };	/* low, high */

	while ((ch = getopt(argc, argv, "df:no:pr:t")) != -1) {
		switch (ch) {
		case 'd':
			debug = 1;
			break;
		case 'f':
			device = optarg;
			break;
		case 'o':
			outputdir = optarg;
			break;
		case 'p':
			psx = 1;
			break;
		case 'n':
			numonly = 1;
			break;
		case 'r':
			if (getrange(optarg, trackrange) != 0)
				errx(1, "invalid range");
			if (debug)
				printf("track range: %u-%u\n", trackrange[0],
				    trackrange[1]);
			break;
		case 't':
			toconly = 1;
			break;
		default:
			errx(1, "usage:	read_cd [-dnpt] [-f device] "
			    "[-o outputdir] [-r single | low-high ]");
		}
	}

	fd = open(device, O_RDONLY);
	if (fd == -1)
		err(1, "open(\"%s\", O_RDONLY)", device);

	if (read_cd(fd, &tracks, numonly) != 0)
		errx(1, "read_cd failure");
	if (toconly || numonly)
		exit(0);
	if (debug)
		tracks->flags |= TR_DEBUG;
	if (psx)
		tracks->flags |= TR_PSX;
	tracks->rangelow = trackrange[0];
	tracks->rangehigh = trackrange[1];
	if (write_tracks(fd, outputdir, tracks) != 0)
		errx(1, "write_tracks failure");
	exit(0);
}

int
read_cd(int fd, struct tracks **rettracks, int numonly) {
	struct tracks *tracks;
	unsigned int track;

	tracks = read_tracks(fd);
	if (tracks == NULL) {
		warn("read_tracks failure");
		return (1);
	}

	if (numonly) {
		printf("%d\n", tracks->tracks[tracks->len - 1].number);
		return (0);
	}

	printf("%5.5s    %10.10s %10.10s %10.10s\n", "Track", "Data type",
	    "Block", "Length");
	printf("%5.5s    %10.10s %10.10s %10.10s\n", "-----", "---------",
	    "-----", "------");
	for (track = 0; track <= tracks->len; track++)
		printf("%5d    %5.5s (%2d) %10d %10d\n",
		    tracks->tracks[track].number,
		    (tracks->tracks[track].type & 4) != 0 ? "data" : "audio",
		    tracks->tracks[track].type,
		    tracks->tracks[track].lba,
		    tracks->tracks[track].len);
	*rettracks = tracks;
	return (0);
}

int
getrange(char *s, unsigned int range[2]) {
	char *endptr, *endptr2;
	unsigned long value = 0;

	if (*s == '-') {
		endptr = s;
		if (endptr[1] == '\0')
			return (-1);
	} else {
		errno = 0;
		value = strtoul(s, &endptr, 0);
		if (errno != 0)
			return (-1);
	}
	switch (*endptr) {
	case '\0':
		/* handle "single" */
		range[0] = range[1] = (unsigned int)value;
		return (0);
	case '-':
		endptr++;
		/* handle "from-" */
		if (*endptr == '\0') {
			range[0] = (unsigned int)value;
			return (0);
		}
		if (endptr - 1 == s) {
			/* handle "-until" */
			errno = 0;
			value = strtoul(endptr, &endptr2, 0);
			if (errno != 0 || *endptr2 != '\0')
				return (-1);
			range[1] = (unsigned int)value;
			return (0);
		}
		/* handle "from-until" */
		if (endptr[0] == '-')
			return (-1);
		range[0] = (unsigned int)value;
		errno = 0;
		value = strtoul(endptr, &endptr2, 0);
		if (errno != 0 || *endptr2 != '\0')
			return (-1);
		range[1] = (unsigned int)value;
		return (0);
	default:
		return (-1);
	}
}
