/*-
 * Copyright (c) 2000 Brian Fundakowski Feldman
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $FreeBSD$
 *
 */

#include <sys/param.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/cdio.h>
#include <sys/cdrio.h>

#include <err.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "tracks.h"

static void
print_offset(const char *name, int fd) {
	off_t offset;

	offset = lseek(fd, (off_t)0, SEEK_CUR);
	if (offset == -1)
		warn("seek on fd %d (%s)", fd, name);
	else
		printf("debug: fd %d (%s) offset %lld\n", fd, name, offset);
}
		

struct tracks *
read_tracks(int fd) {
	struct tracks *tracks;
	struct track *trackarray;
	struct cd_toc_entry *tocarray;
	struct ioc_read_toc_entry entries;
	struct ioc_toc_header toc;
	unsigned int i;

	tracks = calloc(1, sizeof(struct tracks));
	if (tracks == NULL)
		return (NULL);
	if (ioctl(fd, CDIOREADTOCHEADER, &toc) == -1) {
		warn("ioctl(CDIOREADTOCHEADER)");
		free(tracks);
		return (NULL);
	}
	tracks->len = toc.ending_track + 1 - toc.starting_track;
	trackarray = malloc(sizeof(struct track) * (tracks->len + 1));
	if (trackarray == NULL) {
		free(tracks);
		return (NULL);
	}
	tracks->tracks = trackarray;
	tocarray = malloc(sizeof(struct cd_toc_entry) * (tracks->len + 1));
	if (tocarray == NULL) {
		free(tracks->tracks);
		free(tracks);
		return (NULL);
	}
	entries.data = tocarray;
	entries.address_format = CD_LBA_FORMAT;
	entries.starting_track = 0;
	entries.data_len = sizeof(struct cd_toc_entry) * (tracks->len + 1);
	if (ioctl(fd, CDIOREADTOCENTRYS, &entries) == -1) {
		warn("ioctl(CDIOREADTOCENTRYS)");
		free(entries.data);
		free(tracks->tracks);
		free(tracks);
		return (NULL);
	}

	for (i = 0; i <= tracks->len; i++) {
		trackarray[i].type = tocarray[i].control;
		trackarray[i].number = tocarray[i].track;
		trackarray[i].lba = ntohl(tocarray[i].addr.lba);
		trackarray[i].len = ntohl((u_long)tocarray[i + 1].addr.lba) -
		    ntohl((u_long)tocarray[i].addr.lba);
	}
	trackarray[tracks->len].len = 0;
	free(tocarray);
	return (tracks);
}

static int
write_psx(int fd, const char *outputdir, struct track *track, int debug) {
	char *cmd;
	int remainder, rounded;
	int32_t bs = 2048;
	const int spt = 16;

	if (ioctl(fd, CDRIOCSETBLOCKSIZE, &bs) == -1) {
		warn("ioctl(CDRIOCSETBLOCKSIZE)");
		return (1);
	}
	if (lseek(0, (off_t)bs * track->lba, SEEK_SET) == -1) {
		warn("lseek");
		return (1);
	}
	printf("Block size %d, LBA offset %d, length %d, type %d.\n",
	    bs, track->lba, track->len, track->type);
	if (track->len <= 16 + spt * 2) {
		warnx("Length %d way too small for a PSX track.", track->len);
		return (1);
	}
	remainder = (track->len - 2 * spt) % spt;
	rounded = (track->len - 2 * spt - remainder) / spt;
	if (debug)
		printf("Trying read of %d sectors, %d left over.\n",
		    (2 + rounded) * spt, remainder);
	if (debug)
		print_offset("cd", 0);
	/* Get the first 12 blocks. */
	if (asprintf(&cmd,
	    "/bin/dd of=\"./%s/track-%d.%s\" bs=%d count=12",
	    outputdir, track->number,
	    (track->type & 4) != 0 ? "data" : "audio", bs) == -1) {
		warn("asprintf");
		return (1);
	}
	if (system(cmd) != 0) {
		warnx("/bin/dd failure");
		free(cmd);
		return (1);
	}
	free(cmd);
	/* Skip the next 4 (bad) blocks, and mop up til next multiple of spt. */
	if (lseek(0, (off_t)bs * (track->lba + 16), SEEK_SET) == -1) {
		warn("lseek");
		return (1);
	}
	if (debug)
		print_offset("cd", 0);
	if (asprintf(&cmd,
	    "/bin/dd of=\"./%s/track-%d.%s\" bs=%d count=%d seek=%d",
	    outputdir, track->number, (track->type & 4) != 0 ? "data" : "audio",
	    bs, (spt * 2) - 16, 16) == -1) {
		warn("asprintf");
		return (1);
	}
	if (system(cmd) != 0) {
		warnx("/bin/dd failure");
		free(cmd);
		return (1);
	}
	free(cmd);
	/* Read until we get to the end, or.... */
	if (lseek(0, (off_t)bs * (track->lba + 2 * spt), SEEK_SET) == -1) {
		warn("lseek");
		return (1);
	}
	if (debug)
		print_offset("cd", 0);
	if (asprintf(&cmd,
	    "/bin/dd of=\"./%s/track-%d.%s\" ibs=%d obs=%d count=%d "
	    "conv=osync seek=%d",
	    outputdir, track->number,
	    (track->type & 4) != 0 ? "data" : "audio",
	    bs * spt, bs * spt, rounded, 2) == -1) {
		warn("asprintf");
		return (1);
	}
	if (system(cmd) != 0) {
		warnx("/bin/dd failure");
		free(cmd);
		return (1);
	}
	free(cmd);

	if (remainder == 0)
		return (0);

	/* Get the remainder left over */
	if (lseek(0, (off_t)bs * (track->lba + (2 + rounded) * spt),
	    SEEK_SET) == -1) {
		warn("lseek");
		return (1);
	}
	if (debug)
		print_offset("cd", 0);
	if (asprintf(&cmd,
	    "/bin/dd of=\"./%s/track-%d.%s\" ibs=%d obs=%d "
	    "count=%d conv=osync seek=%d",
	    outputdir, track->number, (track->type & 4) != 0 ? "data" : "audio",
	    bs, bs, remainder, (2 + rounded) * spt) == -1) {
		warn("asprintf");
		return (1);
	}
	if (system(cmd) != 0) {
		warnx("/bin/dd failure");
		free(cmd);
		return (1);
	}
	free(cmd);

	return (0);
}

int
write_tracks(int fd, const char *outputdir, struct tracks *tracks) {
	struct stat sb;
	int debug, psx, stdinfd;
	unsigned int tracki;

	debug = tracks->flags & TR_DEBUG;
	psx = tracks->flags & TR_PSX;

restat:
	if (stat(outputdir, &sb) == -1) {
		switch (errno) {
		case ENOENT:
			if (mkdir(outputdir, (mode_t)0700) == -1) {
				warn(outputdir);
				return (1);
			}
			goto restat;
		default:
			warn(outputdir);
			return (1);
		}
	}
	if (!S_ISDIR(sb.st_mode)) {
		warn("\"%s\" is not a directory", outputdir);
		return (1);
	}

	stdinfd = dup(0);
	if (stdinfd == -1) {
		warn("dup(0)");
		return (1);
	}
	if (close(0) == -1) {
		warn("close(0)");
		return (1);
	}
	if (dup2(fd, 0) == -1) {
		warn("dup2(%d, 0)", fd);
		return (1);
	}

	for (tracki = 0; tracki <= tracks->len; tracki++) {
		struct track *track = &tracks->tracks[tracki];
		char *cmd;
		int32_t bs;
		int remainder, rounded, spt;

		if (track->number == 170)
			break;
		if (track->number < tracks->rangelow ||
		    track->number > tracks->rangehigh)
			continue;
		printf("Writing out track %d, ", track->number);
		if ((track->type & 4) != 0) {
			if (psx && track->number == 1) {
				if (write_psx(fd, outputdir, track, debug) == 0)
					continue;
				warnx("write_psx failed");
				return (1);
			}
			bs = 2048;
			spt = 16;
		} else {
			bs = 2352;
			spt = 13;
		}
		if (ioctl(fd, CDRIOCSETBLOCKSIZE, &bs) == -1) {
			warn("ioctl(CDRIOCSETBLOCKSIZE)");
			return (1);
		}
		if (lseek(0, (off_t)bs * track->lba, SEEK_SET) == -1) {
			warn("lseek");
			return (1);
		}
		printf("Block size %d, LBA offset %d, length %d, type %d.\n",
		    bs, track->lba, track->len, track->type);
		remainder = track->len % spt;
		rounded = (track->len - remainder) / spt;
		if (debug)
			printf("Trying read of %d sectors, %d left over.\n",
			    rounded * spt, remainder);
		if (asprintf(&cmd,
		    "/bin/dd of=\"./%s/track-%d.%s\" ibs=%d obs=%d count=%d "
		    "conv=osync",
		    outputdir, track->number,
		    (track->type & 4) != 0 ? "data" : "audio",
		    bs * spt, bs * spt, rounded) == -1) {
			warn("asprintf");
			return (1);
		}
		if (system(cmd) != 0) {
			warnx("/bin/dd failure");
			free(cmd);
			return (1);
		}
		free(cmd);

		if (remainder == 0)
			continue;
		if (lseek(0, (off_t)bs * (track->lba + rounded * spt),
		    SEEK_SET) == -1) {
			warn("lseek");
			return (1);
		}
		if (debug)
			print_offset("cd", 0);
		if (asprintf(&cmd,
		    "/bin/dd of=\"./%s/track-%d.%s\" ibs=%d obs=%d "
		    "count=%d conv=osync seek=%d",
		    outputdir, track->number,
		    (track->type & 4) != 0 ? "data" : "audio",
		    bs, bs, remainder, rounded * spt) == -1) {
			warn("asprintf");
			return (1);
		}
		if (system(cmd) != 0) {
			warnx("/bin/dd failure");
			free(cmd);
			return (1);
		}
		free(cmd);
	}
	if (close(0) == -1) {
		warn("close(0)");
		return (1);
	}
	if (dup2(stdinfd, 0) == -1) {
		warn("dup2(%d, 0)", stdinfd);
		return (1);
	}
	if (close(stdinfd) == -1) {
		warn("close(0)");
		return (1);
	}
	return (0);
}
